--## Choco ##--

Choco = Choco or {util={}}

--[[Converts a string to a table, breaking the string up at every occurance of the separation pattern
* (string) str = String to convert to a table
* (string) sep = Separation pattern to split up string
* (boolean) [plain] = if true then ignore pattern matching for sep (optional)
* (boolean) [reverse] = if true then insert substrings into table in reverse order (optional) ]]
local function str2tbl(str, sep, plain, reverse)
	local version = 1.0
	local date = '2010-06-09'
	if str=='#VERSION' and sep==nil then return version end
	
	local tbl = {} --table to return
	if type(str)=='string' then
		if type(sep)=='string' and sep:len()>=1 then
			if reverse then str = str:reverse() end
			
			local i,n = 1,nil
			while i and i<=str:len() do
				
				local start = i --first index of substring
				n,i = str:find(sep, start, plain) --next occurance of sep
				local stop = n and n-1 --last index of substring
				
				local sub = str:sub(start, stop)
				if reverse then sub = sub and sub:reverse() end --undo reversal
				table.insert(tbl, sub)
				
				i = i and i+1 --index to begin next substring
			end
			
			return tbl
		else return nil end
	else return nil end
end
if str2tbl('#VERSION') > (Choco.util.str2tbl and Choco.util.str2tbl('#VERSION') or 0) then Choco.util.str2tbl=str2tbl end

--[[Convert HH:MM:SS notation to msec
* (string) data = DD:HH:MM:SS or NN (use default units)
* (table) data = {SS,MM,HH,DD}
* (string) default = Units to default to if only one number given (e.g. days, hours, minutes, seconds)
* (number) default = Index to default to if only one number given: 1=seconds, 2=minutes, etc.
-> (number) total_time = sum of data in msec
** Requires: str2tbl ** ]]
local function parseTime(data, default)
	local version = 1.01
	local date = '2011-02-03'
	if data=='#VERSION' then return version end
	
	local time_tbl
	if type(data)=='string' then time_tbl = str2tbl(data, ':', nil, true)
	elseif type(data)=='table' then time_tbl = data
	else return end
	
	local total_time = 0 --in msecs
	local scale = {1000, 60, 60, 24} --multiplier to convert to msec
	local defaults = {seconds=1, minutes=2, hours=3, days=4} --convert default as string to index
	local mult = 1 --cumulative multiplier
	default = type(default)=='string' and (defaults[default] or 0) or math.min(math.floor(tonumber(default) or 0), table.maxn(scale) or 0)
	
	local default_mult = 1
	if default >= 1 then
		for k=1,default do default_mult=default_mult*scale[k] end
	end
	
	if table.maxn(time_tbl)==1 and tonumber(time_tbl[1]) then return default_mult*tonumber(time_tbl[1]) end --return single entry as minutes
	
	for i,v in ipairs(time_tbl) do
		if scale[i] then
			local val = tonumber(v)
			mult = mult*scale[i]
			total_time = total_time + (val and mult*val or 0)
		end
	end
	
	return total_time
end
if parseTime('#VERSION') > (Choco.util.parseTime and Choco.util.parseTime('#VERSION') or 0) then Choco.util.parseTime=parseTime end

--[[Convert msec to table containing days, hours, minutes, seconds
* (number) msec = value in milliseconds
-> (table) time = {seconds=SS,minutes=MM,hours=HH,days=DD,total=msec} ]]
local function msec2tbl(msec)
	local version = 1.0
	local date = '2010-06-11'
	if msec=='#VERSION' then return version end
	
	msec = tonumber(msec) or 0
	local time = {days=0, hours=0, minutes=0, seconds=0, total=msec}
	local const = 1000*60*60*24
	
	if msec >= const then
		time.days = math.floor(msec/const)
		msec = msec%const
	end
	const = 1000*60*60
	if msec >= const then
		time.hours = math.floor(msec/const)
		msec = msec%const
	end
	const = 1000*60
	if msec >= const then
		time.minutes = math.floor(msec/const)
		msec = msec%const
	end
	time.seconds = msec/1000
	
	return time
end
if msec2tbl('#VERSION') > (Choco.util.msec2tbl and Choco.util.msec2tbl('#VERSION') or 0) then Choco.util.msec2tbl=msec2tbl end

--[[Converts msec to string giving days, hours, minutes, seconds
*  (number) msec = value in milliseconds
-> (string) str = [DD days ][HH hours ][MM minutes ][SS seconds]
** Requires: msec2tbl ** ]]
local function msec2str(msec)
	local version = 1.0
	local date = '2010-06-11'
	if msec=='#VERSION' then return version end
	
	local time = msec2tbl(msec) or {}
	local str = ''
	
	if time.days and time.days>0 then str = str..time.days..' days ' end
	if time.hours and time.hours>0 then str = str..time.hours..' hours ' end
	if time.minutes and time.minutes>0 then str = str..time.minutes..' minutes ' end
	if time.seconds and time.seconds>0 or time.total==0 then str = str..time.seconds..' seconds ' end
	return (str)
end
if msec2str('#VERSION') > (Choco.util.msec2str and Choco.util.msec2str('#VERSION') or 0) then Choco.util.msec2str=msec2str end