countdowntimer = {init={},event={}}
countdowntimer.main = nil --main timer (can get time elapsed/remaining)
countdowntimer.timers = {} --additional bg timers
countdowntimer.version = '1.3.2'

--## Choco ##--

dofile'choco.lua'
countdowntimer.vars = dofile'vars.lua'
countdowntimer.Timer = dofile'timer.lua'

local chat_outputs = {
	sector = SendChat,
	group = SendChat,
	hud = function(msg) if msg then HUD:PrintSecondaryMsg('\12700ff00'..msg) end end,
	print = function(msg) if msg then print('\127ffffff'..msg) end end,
	none = function() end, --don't print any messages
}

--Check timestamp saved to config file to see if timer needs to be resumed
function countdowntimer.init:PLAYER_ENTERED_GAME(event, data)
	UnregisterEvent(self, "PLAYER_ENTERED_GAME") --unregister because don't need to re-init on character change
	
	local start_time = gkini.ReadInt("Timer", "start_time", 0)
	local duration = gkini.ReadInt("Timer", "duration", 0)
	
	if start_time>0 and duration>0 then
		local diff_time = os.difftime(os.time(), start_time)*1000 --time since timer last started (msec)
	
		if diff_time < duration then --timer has not yet expired
			countdowntimer.main = countdowntimer.main or countdowntimer.Timer:new()
			countdowntimer.main.start_time = start_time
			countdowntimer.main.duration = duration
			
			local remaining = duration-diff_time
			countdowntimer.main.clock:SetTimeout(remaining, function() countdowntimer.main:complete() end)
			print('\127ffffffTimer resumed with '..Choco.util.msec2str(remaining)..' remaining')
		end
	end
	
	--countdowntimer.Timer.start_time = start_time --OBSOLETE
	--countdowntimer.Timer.duration = duration --OBSOLETE
end

--[[ Triggered when an incremental timer reaches a milestone
* (table) data = {id=timer.start_time,duration=timer.duration,desc=timer.desc,params=timer.params,pending=timer.pending} ]]
function countdowntimer.event:CHOCO_TIMER_INCREMENT(event, data)
	local output = data.params and data.params.output or 'print'
	chat_outputs[output](Choco.util.msec2str(data.pending), output:upper())
	
	local count = math.floor(data.pending/1000+0.5)
	if count>0 then
		if count <= 3 and countdowntimer.vars.playsound==1 then
			gksound.GKPlaySound('chime'..count)
		else gksound.GKPlaySound'tick' end
	end
end

--[[ Triggered when a timer expires
* (table) data = {id=timer.start_time,duration=timer.duration,desc=timer.desc} ]]
function countdowntimer.event:CHOCO_TIMER_COMPLETED(event, data)
	local msg = ''
	if data.params and data.params.is_increment then
		msg = data.desc or 'GO!'
		if countdowntimer.vars.playsound==1 then gksound.GKPlaySound'chime0' end
	else
		msg = data.desc or countdowntimer.vars.default_msg
		if countdowntimer.vars.playsound==1 then gksound.GKPlaySound'timer_end' end
	end
	
	local output = data.params and data.params.output or 'print'
	chat_outputs[output](msg, output:upper())
	
	--reset master elapsed time if main timer ended
	if data.params and data.params.is_main then
	--if countdowntimer.main and data.id==countdowntimer.main.start_time and not countdowntimer.main.clock:IsActive() then --OBSOLETE
		--TODO need to distinguish if main timer or bg timer completed
		--countdowntimer.main = nil --OBSOLETE
		--countdowntimer.Timer.start_time = gkmisc.GetGameTime() --OBSOLETE
		--countdowntimer.Timer.duration = 0 --OBSOLETE
		
		--save timestamp to return elapsed time after main expired
		gkini.WriteInt("Timer", "start_time", os.time())
		gkini.WriteInt("Timer", "duration", 0)
	end
	
	--remove all inactive timers
	for n=table.maxn(countdowntimer.timers),1,-1 do --reverse loop through timers table
		if countdowntimer.timers[n] and not countdowntimer.timers[n].clock:IsActive() then table.remove(countdowntimer.timers,n) end
	end
end

--Input from console
countdowntimer.Inputs = {set=true,new=true,kill=true,msg=true,elapsed=true,remaining=true,sector=true,group=true,sound=true,help=true}
function countdowntimer:Input(args)
	args = args or {}
	local command = args[1] and args[1]:lower()
	if command then table.remove(args, 1) end
	
	if countdowntimer.Inputs[command] then
		countdowntimer[command](countdowntimer,args)
	else countdowntimer.help(countdowntimer) end
end

--Start new background timer
function countdowntimer:new(args)
	local new_timer = self.Timer:new(args)
	table.insert(self.timers, new_timer)
	if new_timer.duration then
		local group_warning = new_timer.params and new_timer.params.output=='group' and not IsGroupMember(GetCharacterID()) and '. You must be in a group to use group chat!' or ''
		print('\127ffffffTimer for '..Choco.util.msec2str(new_timer.duration)..' has been set'..group_warning)
	else print'\127ffffffYou must specify the timer duration, either HH:MM:SS or just minutes' end
end

--Activate main timer
function countdowntimer:set(args)
	self.main = self.main or self.Timer:new()
	self.main:msg(self.vars.default_msg)
	local msec = self.main:set(args) 
	if msec then
		--save data in case of interrupt
		gkini.WriteInt("Timer", "start_time", os.time())
		gkini.WriteInt("Timer", "duration", msec)
		
		self.main.params = self.main.params or {}
		self.main.params.is_main = true
		
		local group_warning = self.main.params and self.main.params.output=='group' and not IsGroupMember(GetCharacterID()) and '. You must be in a group to use group chat!' or ''
		print('\127ffffffTimer for '..Choco.util.msec2str(msec)..' has been set'..group_warning)
	else print'\127ffffffYou must specify the timer duration, either HH:MM:SS or just minutes' end
end

--Kill active timer(s)
function countdowntimer:kill(args)
	args = args or {}
	
	local success_main
	local success_bg
	if args[1]=='all' or args[1]=='bg' then
		for n=table.maxn(self.timers),1,-1 do
			if self.timers[n] and self.timers[n]:kill() then
				table.remove(self.timers, n)
				success_bg = true
			end
		end
		success_bg = success_bg or false
	elseif args[1] then self:help{'kill'}; return end --invalid argument
	
	if args[1]~='bg' then success_main = self.main and self.main:kill() end
	
	local msg
	if success_bg==nil then
		msg = success_main and 'Main timer aborted' or 'Main timer has not been set' --TODO give number of active bg timers
	elseif success_main==nil then
		msg = success_bg and 'Background timers aborted' or 'No background timers are set'
	else msg = (not success_main and not success_bg) and 'No timers have been set' or 'All timers aborted' end
	
	print('\127ffffff'..msg)
	
	if success_main then
		gkini.WriteInt("Timer", "start_time", os.time())
		gkini.WriteInt("Timer", "duration", 0)
		
		--self.main = nil --OBSOLETE
		--print'\127ffffffTimer aborted' --OBSOLETE
	--else print'\127ffffffTimer has not been set' end --OBSOLETE
	end
end

--Set message to display when main timer expires
function countdowntimer:msg(args)
	if self.main then self.main:msg(args) end --change message if and for active timer
	local msg = args and args[1] and table.concat(args, " ")
	if msg then
		self.vars.default_msg = msg --change message for all new timers
		print('\127ffffffTimer msg set to: '..msg)
	else self:help{'msg'} end
end

--Print elapsed time for main timer (since started or since expired)
function countdowntimer:elapsed(args)
	local timer = self.main or {start_time=gkini.ReadInt("Timer", "start_time", 0), duration=gkini.ReadInt("Timer", "duration", 0), elapsed=self.Timer.elapsed} --entries needed for timer:elapsed()
	local msec = timer:elapsed(args)
	
	if msec then
		if msec>0 then
			print('\127ffffffTimer elapsed time: '..Choco.util.msec2str(msec))
		else print('\127ffffffTimer expired '..Choco.util.msec2str(-msec)..' ago') end
	else print'\127ffffffTimer has not been set' end
end

--Print remaining time on main timer if active
function countdowntimer:remaining(args)
	local remaining = self.main and self.main:remaining(args)
	if remaining then print('\127ffffffTimer remaining time: '..remaining) else print'\127ffffffTimer has not been set' end
end

--Create new background timer output to sector chat
function countdowntimer:sector(args)
	table.insert(args, 1, 'sector')
	self:new(args)
end

--Create new background timer output to group chat
function countdowntimer:group(args)
	table.insert(args, 1, 'group')
	self:new(args)
end

--Enable or disable sound playback when timer expires
function countdowntimer:sound(args)
	args = args or {}
	local param = args[1] and args[1]:lower()
	local inputs = {on=1,off=0,toggle=(self.vars.playsound+1)%2,yes=1,no=0,['true']=1,['false']=0} --keys are valid entries for param
	
	if inputs[param] then
		self.vars.playsound = inputs[param]
		gkini.WriteInt("Timer", "sound", self.vars.playsound) --save to config
		print('\127ffffffTimer sound playback is '..(self.vars.playsound==1 and 'ON' or 'OFF'))
	else self:help{'sound'} end --invalid parameters, display help for sound command
end

--Text to display for help command
function countdowntimer:help(args)
	args = args or {}
	
	local help_text = {
		['help']='\127ffffffTimer v'..self.version..'  Available commands are: (for more info use /timer help \127ffff00COMMAND\127o)'..
			'\n\t/timer set \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Starts main timer with countdown for \1270070c0HH\127o hours, \1270070c0MM\127o minutes, and \1270070c0SS\127o seconds (or just \1270070c0MM\127o minutes)'..
			'\n\t/timer new \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Starts a background timer (syntax identical to /timer set). Elapsed and remaining commands only work on main timer'..
			'\n\t/timer sector/group \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Creates a new timer that outputs to sector/group chat incrementally'.. 
			'\n\t/timer elapsed OR /timer_elapsed - Display time since the main timer was activated (or expired)'..
			'\n\t/timer remaining OR /timer_remaining - Display remaining time until the main timer expires'..
			'\n\t/timer msg \127ffff00TEXT\127o - Where \127ffff00TEXT\127o is displayed once main timer completes'..
			'\n\t/timer kill - aborts main timer if active. To kill background timers too use: /timer kill all'..
			'\n\t/timer sound \12700ff00on\127o/\12700ff00off\127o/\12700ff00toggle\127o - Plays sound when timer completes if set to ON',
		['set']='\127ffffff/timer set \127a0a0a0[\12700ff00sector\127o/\12700ff00group\127o/\12700ff00hud\127o/\12700ff00none\127o] [\12700ffffincrement=\127o\12700ff00yes\127o/\12700ff00no\127o]\127ffffff \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o \127a0a0a0[\127ffff00TEXT\127o]\127ffffff - Starts main timer'..
			'\n\t\12700ff00sector\127o/\12700ff00group\127o/\12700ff00hud\127o/\12700ff00none\127o - \127a0a0a0(optional)\127o output instead to sector or group chat or HUD. Audio only if none: /timer set sector 5'..
			'\n\t\12700ffffincrement=\127o\12700ff00yes\127o/\12700ff00no\127o - \127a0a0a0(optional)\127o output incremental updates, default=yes for sector/group: /timer set increment=yes 5:30'..
			'\n\t\1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Sets countdown for \1270070c0HH\127o hours, \1270070c0MM\127o minutes, and \1270070c0SS\127o seconds (or just \1270070c0MM\127o minutes; just \1270070c0SS\127o seconds if sector or group)'..
			'\n\t   Decimals are also allowed. For half an hour: /timer set 0.5:0:0'..
			'\n\t\127ffff00TEXT\127o - \127a0a0a0(optional)\127o message to display when timer completes: /timer set 1:30 "Time\'s up!"'..
			'\n\tFor a timer that counts up indefinately use: /timer set 0',
		['new']='\127ffffff/timer new \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Starts a new background timer with countdown for \1270070c0HH\127o hours, \1270070c0MM\127o minutes, and \1270070c0SS\127o seconds (or just \1270070c0MM\127o minutes)'..
			'\n\tSame as set command except creates a new timer in the background instead of using the main timer'..
			'\n\tSee Also: /timer help set',
		['sector']='\127ffffff/timer sector \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Creates a new timer that outputs to sector chat incrementally'..
			'\n\tEquivalent to: /timer new sector \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o',
		['group']='\127ffffff/timer group \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o - Creates a new timer that outputs to group chat incrementally'..
			'\n\tEquivalent to: /timer new group \1270070c0HH\127o:\1270070c0MM\127o:\1270070c0SS\127o',	
		['elapsed']='\127ffffff/timer elapsed - Display time since main timer was activated (or expired)'..
			'\n\tTo create a bind for quick access use: /bind KEY timer_elapsed',
		['remaining']='\127ffffff/timer remaining - Display remaining time until main timer expires'..
			'\n\tTo create a bind for quick access use: /bind KEY timer_remaining',
		['msg']='\127ffffff/timer msg \127ffff00TEXT\127o - Where \127ffff00TEXT\127o is displayed once main timer completes'..
			'\n\tNote that the message used is not saved to the config file, but will be used for all future main timers this session'..
			'\n\tMessage is currently set to: \127a0a0a0'..(self.main and self.main.desc or self.vars.default_msg)..'\127o',
		['kill']='\127ffffff/timer kill - aborts main timer if active'..
			'\n\tTo kill all active timers use: /timer kill all'..
			'\n\tTo kill all timers except the main timer use: /timer kill bg'..
			'\n\tNumber of active timers: '..self.numActiveTimers(),
		['sound']='\127ffffff/timer sound \12700ff00on\127o/\12700ff00off\127o/\12700ff00toggle\127o - Plays sound when timer completes if set to ON (saved to config)'..
			'\n\tSound playback is set to: \127a0a0a0'..(self.vars.playsound==1 and 'ON' or 'OFF')..'\127o', --TODO vars func to get playsound vars
		['command']='\127ffffffTo get more information about how to use a command, type: /timer help \127ffff00COMMAND\127o'..
			'\n\tWhere \127ffff00COMMAND\127o is the command you want to get help on (e.g. /timer help set)',
	}
	print(help_text[ args[1] ] or help_text.help)
end

--returns number of active timers (main & bg timers)
function countdowntimer.numActiveTimers()
	local count = 0
	if countdowntimer.main and countdowntimer.main.clock:IsActive() then count = count + 1 end
	for _,v in ipairs(countdowntimer.timers) do
		if v.clock:IsActive() then count = count+1 end
	end
	
	return count
end

RegisterEvent(countdowntimer.init, "PLAYER_ENTERED_GAME")
RegisterEvent(countdowntimer.event, "CHOCO_TIMER_INCREMENT")
RegisterEvent(countdowntimer.event, "CHOCO_TIMER_COMPLETED")
RegisterUserCommand("timer", countdowntimer.Input)
RegisterUserCommand("timer_elapsed", function() countdowntimer:Input{'elapsed'} end) --convenience for binds
RegisterUserCommand("timer_remaining", function() countdowntimer:Input{'remaining'} end) --convenience for binds